==> pyTermTk-0.12.0a4-py3-none-any.whl: TermTk/TTkWidgets/splitter.py <==
#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2021 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from TermTk.TTkCore.constant import TTkK
from TermTk.TTkWidgets.widget import *
from TermTk.TTkWidgets.frame import *

class TTkSplitter(TTkFrame):
    __slots__ = (
        '_splitterInitialized', '_orientation',
        '_separators', '_separatorsRef', '_sizeRef', '_initSizes',
        '_widgets', '_separatorSelected')
    def __init__(self, *args, **kwargs):
        self._splitterInitialized = False

        self._widgets = []
        self._separators = []
        self._separatorsRef = []
        self._sizeRef = 0
        self._initSizes = []
        self._separatorSelected = None
        self._orientation = TTkK.HORIZONTAL
        TTkFrame.__init__(self, *args, **kwargs)
        self._name = kwargs.get('name' , 'TTkSpacer')
        self._orientation = kwargs.get('orientation', TTkK.HORIZONTAL)
        self.setBorder(kwargs.get('border' , False))
        self.setFocusPolicy(TTkK.ClickFocus)
        self._splitterInitialized = True

        class _SplitterLayout(TTkLayout):
            def insertWidget(_, index, widget):
                self.insertWidget(index, widget)
            def addWidget(_, widget):
                self.addWidget(widget)
        self.setLayout(_SplitterLayout())

    def orientation(self):
        return self._orientation

    def count(self):
        return len(self._widgets)

    def indexOf(self, widget):
        return self._widgets.index(widget)

    def widget(self, index):
        return self._widgets[index]

    def replaceWidget(self, index, widget):
        TTkLayout.removeWidget(self.layout(), self._widgets[index])
        TTkLayout.insertWidget(self.layout(), index, widget)
        self._widgets[index] = widget
        self._updateGeometries()

    def removeWidget(self, widget):
        index = self.indexOf(widget)
        self._widgets.pop(index)
        self._initSizes.pop(index)
        self._separators.pop(index)
        self._separatorsRef = [s for s in self._separators]
        TTkLayout.removeWidget(self.layout(), widget)
        self._updateGeometries()

    def addWidget(self, widget, size=None):
        self.insertWidget(len(self._widgets), widget, size)

    def insertWidget(self, index, widget, size=None):
        TTkLayout.insertWidget(self.layout(), index, widget)
        _,_,w,h = self.geometry()
        if self.border():
            w-=2
            h-=2
        self._widgets.insert(index, widget)
        numW = len(self._widgets)

        if self._orientation == TTkK.HORIZONTAL:
            fullSize = w
        else:
            fullSize = h
        # assign the same slice to all the widgets
        self._initSizes.insert(index, size)
        self._separators = [fullSize*i//numW for i in range(1,numW+1)]
        self._updateGeometries()
        self._separatorsRef = self._separators
        self._sizeRef = fullSize
        if self.parentWidget():
            self.parentWidget().update(repaint=True, updateLayout=True)

    def setSizes(self, sizes):
        s = 0
        sizes=sizes[:len(self._separators)]
        for i, l in enumerate(sizes):
            s += l
            self._separators[i] = s
            self._separatorsRef[i] = s
            self._initSizes[i] = l
        self._updateGeometries()


    def _minMaxSizeBefore(self, index):
        if self._separatorSelected is None:
            return 0, 0x1000
        # this is because there is a hidden splitter at position -1
        minsize = -1
        maxsize = -1
        for i in range(self._separatorSelected+1):
            item = self._widgets[i]
            minsize += item.minDimension(self._orientation)+1
            maxsize += item.maxDimension(self._orientation)+1
        return minsize, maxsize

    def _minMaxSizeAfter(self, index):
        if self._separatorSelected is None:
            return 0, 0x1000
        minsize = 0x0
        maxsize = 0x0
        for i in range(self._separatorSelected+1, len(self._separators)):
            item = self._widgets[i]
            minsize += item.minDimension(self._orientation)+1
            maxsize += item.maxDimension(self._orientation)+1
        return minsize, maxsize

    def _updateGeometries(self, resized=False):
        if not self.isVisible(): return
        _,_,w,h = self.geometry()
        sep = self._separators = self._separators[0:len(self._widgets)]
        if self.border():
            w-=2
            h-=2

        def _processGeometry(index, forward):
            item = self._widgets[index]
            pa = -1 if index==0 else sep[index-1]
            pb = sep[index]

            if self._orientation == TTkK.HORIZONTAL:
                newPos = pa+1
                size = w-newPos
            else:
                newPos = pa+1
                size = h-newPos

            if i<=len(sep)-2: # this is not the last widget
                size = pb-newPos
                maxsize = item.maxDimension(self._orientation)
                minsize = item.minDimension(self._orientation)
                if   size > maxsize: size = maxsize
                elif size < minsize: size = minsize
                if forward:
                    sep[index]=pa+size+1
                elif i>0 :
                    sep[index-1]=pa=pb-size-1

            if self._orientation == TTkK.HORIZONTAL:
                item.setGeometry(pa+1,0,size,h)
            else:
                item.setGeometry(0,pa+1,w,size)
            pass


        selected = 0
        if self._orientation == TTkK.HORIZONTAL:
            size = w
        else:
            size = h
        if self._separatorSelected is not None:
            selected = self._separatorSelected
            sepPos = sep[selected]
            minsize,maxsize = self._minMaxSizeBefore(selected)
            # TTkLog.debug(f"before:{minsize,maxsize}")
            if sepPos > maxsize: sep[selected] = maxsize
            if sepPos < minsize: sep[selected] = minsize
            minsize,maxsize = self._minMaxSizeAfter(selected)
            # TTkLog.debug(f"after:{minsize,maxsize}")
            if sepPos < size-maxsize: sep[selected] = size-maxsize
            if sepPos > size-minsize: sep[selected] = size-minsize

        if resized:
            l = len(sep)
            for i in reversed(range(l)):
                _processGeometry(i, False)
            for i in range(l):
                _processGeometry(i, True)
        else:
            for i in reversed(range(selected+1)):
                _processGeometry(i, False)
            for i in range(selected+1, len(sep)):
                _processGeometry(i, True)

        if self._separatorSelected is not None or self._sizeRef==0:
            self._separatorsRef = self._separators
            self._sizeRef = size

    def resizeEvent(self, w, h):
        if w==h==0: return
        if not self._sizeRef:
            # This is the first resize (w,h != 0 and previous reference size was 0)
            # I need to define the initial position of all the widgets
            if self._orientation == TTkK.HORIZONTAL:
                self._sizeRef = w
            else:
                self._sizeRef = h
            # get the sum of the fixed sizes
            fixSize = sum(filter(None, self._initSizes))
            numVarSizes = len([x for x in self._initSizes if x is None])
            avalSize = self._sizeRef-fixSize
            sizes = [avalSize//numVarSizes if s is None else s for s in self._initSizes]
            self._separatorsRef = [sum(sizes[:i+1]) for i in range(len(sizes))]

        # Adjust separators to the new size;
        self._separatorSelected = None
        if self._sizeRef > 0:
            if self._orientation == TTkK.HORIZONTAL:
                diff = w/self._sizeRef
            else:
                diff = h/self._sizeRef
            self._separators = [int(i*diff) for i in self._separatorsRef]
        self._updateGeometries(resized=True)

    def paintEvent(self):
        off = 1 if self.border() else 0
        TTkFrame.paintEvent(self)
        w,h = self.size()
        if self._orientation == TTkK.HORIZONTAL:
            for i in self._separators[:-1]:
                self._canvas.drawVLine(pos=(i+off,0), size=h)
        else:
            for i in self._separators[:-1]:
                self._canvas.drawHLine(pos=(0,i+off), size=w)

    def mousePressEvent(self, evt):
        self._separatorSelected = None
        x,y = evt.x, evt.y
        if self.border():
            x-=1 ; y-=1
        # TTkLog.debug(f"{self._separators} {evt}")
        for i, val in enumerate(self._separators):
            if self._orientation == TTkK.HORIZONTAL:
                if x == val:
                    self._separatorSelected = i
                    self.update()
                    self._updateGeometries()
            else:
                if y == val:
                    self._separatorSelected = i
                    self.update()
                    self._updateGeometries()
        return self._separatorSelected is not None

    def mouseDragEvent(self, evt):
        if self._separatorSelected is not None:
            x,y = evt.x, evt.y
            if self.border():
                x-=1 ; y-=1
            if self._orientation == TTkK.HORIZONTAL:
                self._separators[self._separatorSelected] = x
            else:
                self._separators[self._separatorSelected] = y
            self._updateGeometries()
            self.update()
            return True
        return False

    def focusOutEvent(self):
        self._separatorSelected = None

    def minimumHeight(self) -> int:
        ret = 2 if self.border() else 0
        if not self._splitterInitialized: return ret
        if self._orientation == TTkK.VERTICAL:
            for item in self._widgets:
                ret+=item.minimumHeight()+1
            ret = max(0,ret-1)
        else:
            for item in self._widgets:
                if ret < item.minimumHeight():
                    ret = item.minimumHeight()
        return ret

    def minimumWidth(self)  -> int:
        ret = 2 if self.border() else 0
        if not self._splitterInitialized: return ret
        if self._orientation == TTkK.HORIZONTAL:
            for item in self._widgets:
                ret+=item.minimumWidth()+1
            ret = max(0,ret-1)
        else:
            for item in self._widgets:
                if ret < item.minimumWidth():
                    ret = item.minimumWidth()
        return ret

    def maximumHeight(self) -> int:
        b = 2 if self.border() else 0
        if not self._splitterInitialized: return 0x10000
        if self._orientation == TTkK.VERTICAL:
            ret = b
            for item in self._widgets:
                ret+=item.maximumHeight()+1
            ret = max(b,ret-1)
        else:
            ret = 0x10000
            for item in self._widgets:
                if ret > item.maximumHeight():
                    ret = item.maximumHeight()
        return ret

    def maximumWidth(self)  -> int:
        b = 2 if self.border() else 0
        if not self._splitterInitialized: return 0x10000
        if self._orientation == TTkK.HORIZONTAL:
            ret = b
            for item in self._widgets:
                ret+=item.maximumHeight()+1
            ret = max(b,ret-1)
        else:
            ret = 0x10000
            for item in self._widgets:
                if ret > item.maximumWidth():
                    ret = item.maximumWidth()
        return ret

==> pyTermTk-0.20.0a0-py3-none-any.whl: TermTk/TTkWidgets/splitter.py <==
#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2021 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from TermTk.TTkCore.constant import TTkK
from TermTk.TTkWidgets.widget import *
from TermTk.TTkWidgets.frame import *

class TTkSplitter(TTkFrame):
    '''TTkSplitter'''
    __slots__ = (
        '_splitterInitialized', '_orientation',
        '_separators', '_separatorsRef', '_sizeRef', '_initSizes',
        '_items', '_separatorSelected')
    def __init__(self, *args, **kwargs):
        self._splitterInitialized = False

        self._items = []
        self._separators = []
        self._separatorsRef = []
        self._sizeRef = 0
        self._initSizes = []
        self._separatorSelected = None
        self._orientation = TTkK.HORIZONTAL
        TTkFrame.__init__(self, *args, **kwargs)
        self._orientation = kwargs.get('orientation', TTkK.HORIZONTAL)
        self.setBorder(kwargs.get('border' , False))
        self.setFocusPolicy(TTkK.ClickFocus)
        self._splitterInitialized = True

        class _SplitterLayout(TTkLayout):
            def insertWidget(_, index, widget):
                self.insertWidget(index, widget)
            def addWidget(_, widget):
                self.addWidget(widget)
            def inserItem(_, item):
                self.inserItem(item)
            def addItem(_, item):
                self.addItem(item)
        self.setLayout(_SplitterLayout())

    def orientation(self):
        '''orientation'''
        return self._orientation

    def count(self):
        '''count'''
        return len(self._items)

    def indexOf(self, widget):
        '''indexOf'''
        return self._items.index(widget)

    def widget(self, index):
        '''widget'''
        return self._items[index]

    def replaceItem(self, index, item):
        '''replaceItem'''
        TTkLayout.removeItem(self.layout(), self._items[index])
        TTkLayout.insertItem(self.layout(), index, item)
        self._items[index] = item
        self._updateGeometries()

    def replaceWidget(self, index, widget):
        '''replaceWidget'''
        TTkLayout.removeWidget(self.layout(), self._items[index])
        TTkLayout.insertWidget(self.layout(), index, widget)
        self._items[index] = widget
        self._updateGeometries()

    def removeItem(self, item):
        '''removeItem'''
        index = self.indexOf(item)
        self._items.pop(index)
        self._initSizes.pop(index)
        self._separators.pop(index)
        self._separatorsRef = [s for s in self._separators]
        TTkLayout.removeItem(self.layout(), item)
        self._updateGeometries()

    def removeWidget(self, widget):
        '''removeWidget'''
        index = self.indexOf(widget)
        self._items.pop(index)
        self._initSizes.pop(index)
        self._separators.pop(index)
        self._separatorsRef = [s for s in self._separators]
        TTkLayout.removeWidget(self.layout(), widget)
        self._updateGeometries()

    def addItem(self, item, size=None):
        '''addItem'''
        self.insertItem(len(self._items), item, size)

    def insertItem(self, index, item, size=None):
        '''insertItem'''
        TTkLayout.insertItem(self.layout(), index, item)
        self._insertWidgetItem(index, item, size)

    def addWidget(self, widget, size=None):
        '''addWidget'''
        self.insertWidget(len(self._items), widget, size)

    def insertWidget(self, index, widget, size=None):
        '''insertWidget'''
        TTkLayout.insertWidget(self.layout(), index, widget)
        self._insertWidgetItem(index, widget, size)

    def _insertWidgetItem(self, index, widgetItem, size=None):
        _,_,w,h = self.geometry()
        if self.border():
            w-=2
            h-=2
        self._items.insert(index, widgetItem)
        numW = len(self._items)

        if self._orientation == TTkK.HORIZONTAL:
            fullSize = w
        else:
            fullSize = h
        # assign the same slice to all the widgets
        self._initSizes.insert(index, size)
        self._separators = [fullSize*i//numW for i in range(1,numW+1)]
        self._updateGeometries()
        self._separatorsRef = self._separators
        self._sizeRef = fullSize
        if self.parentWidget():
            self.parentWidget().update(repaint=True, updateLayout=True)

    def setSizes(self, sizes):
        '''setSizes'''
        s = 0
        sizes=sizes[:len(self._separators)]
        for i, l in enumerate(sizes):
            s += l
            self._separators[i] = s
            self._separatorsRef[i] = s
            self._initSizes[i] = l
        self._updateGeometries()


    def _minMaxSizeBefore(self, index):
        if self._separatorSelected is None:
            return 0, 0x1000
        # this is because there is a hidden splitter at position -1
        minsize = -1
        maxsize = -1
        for i in range(self._separatorSelected+1):
            item = self._items[i]
            minsize += item.minDimension(self._orientation)+1
            maxsize += item.maxDimension(self._orientation)+1
        return minsize, maxsize

    def _minMaxSizeAfter(self, index):
        if self._separatorSelected is None:
            return 0, 0x1000
        minsize = 0x0
        maxsize = 0x0
        for i in range(self._separatorSelected+1, len(self._separators)):
            item = self._items[i]
            minsize += item.minDimension(self._orientation)+1
            maxsize += item.maxDimension(self._orientation)+1
        return minsize, maxsize

    def _updateGeometries(self, resized=False):
        if not self.isVisible(): return
        _,_,w,h = self.geometry()
        sep = self._separators = self._separators[0:len(self._items)]
        if self.border():
            w-=2
            h-=2

        def _processGeometry(index, forward):
            item = self._items[index]
            pa = -1 if index==0 else sep[index-1]
            pb = sep[index]

            if self._orientation == TTkK.HORIZONTAL:
                newPos = pa+1
                size = w-newPos
            else:
                newPos = pa+1
                size = h-newPos

            if i<=len(sep)-2: # this is not the last widget
                size = pb-newPos
                maxsize = item.maxDimension(self._orientation)
                minsize = item.minDimension(self._orientation)
                if   size > maxsize: size = maxsize
                elif size < minsize: size = minsize
                if forward:
                    sep[index]=pa+size+1
                elif i>0 :
                    sep[index-1]=pa=pb-size-1

            if self._orientation == TTkK.HORIZONTAL:
                item.setGeometry(pa+1,0,size,h)
            else:
                item.setGeometry(0,pa+1,w,size)
            pass


        selected = 0
        if self._orientation == TTkK.HORIZONTAL:
            size = w
        else:
            size = h
        if self._separatorSelected is not None:
            selected = self._separatorSelected
            sepPos = sep[selected]
            minsize,maxsize = self._minMaxSizeBefore(selected)
            # TTkLog.debug(f"before:{minsize,maxsize}")
            if sepPos > maxsize: sep[selected] = maxsize
            if sepPos < minsize: sep[selected] = minsize
            minsize,maxsize = self._minMaxSizeAfter(selected)
            # TTkLog.debug(f"after:{minsize,maxsize}")
            if sepPos < size-maxsize: sep[selected] = size-maxsize
            if sepPos > size-minsize: sep[selected] = size-minsize

        if resized:
            l = len(sep)
            for i in reversed(range(l)):
                _processGeometry(i, False)
            for i in range(l):
                _processGeometry(i, True)
        else:
            for i in reversed(range(selected+1)):
                _processGeometry(i, False)
            for i in range(selected+1, len(sep)):
                _processGeometry(i, True)

        if self._separatorSelected is not None or self._sizeRef==0:
            self._separatorsRef = self._separators
            self._sizeRef = size

    def resizeEvent(self, w, h):
        if w==h==0: return
        if not self._sizeRef:
            # This is the first resize (w,h != 0 and previous reference size was 0)
            # I need to define the initial position of all the widgets
            if self._orientation == TTkK.HORIZONTAL:
                self._sizeRef = w
            else:
                self._sizeRef = h
            # get the sum of the fixed sizes
            fixSize = sum(filter(None, self._initSizes))
            numVarSizes = len([x for x in self._initSizes if x is None])
            avalSize = self._sizeRef-fixSize
            sizes = [avalSize//numVarSizes if s is None else s for s in self._initSizes]
            self._separatorsRef = [sum(sizes[:i+1]) for i in range(len(sizes))]

        # Adjust separators to the new size;
        self._separatorSelected = None
        if self._sizeRef > 0:
            if self._orientation == TTkK.HORIZONTAL:
                diff = w/self._sizeRef
            else:
                diff = h/self._sizeRef
            self._separators = [int(i*diff) for i in self._separatorsRef]
        self._updateGeometries(resized=True)

    def paintEvent(self):
        off = 1 if self.border() else 0
        TTkFrame.paintEvent(self)
        w,h = self.size()
        if self._orientation == TTkK.HORIZONTAL:
            for i in self._separators[:-1]:
                self._canvas.drawVLine(pos=(i+off,0), size=h)
        else:
            for i in self._separators[:-1]:
                self._canvas.drawHLine(pos=(0,i+off), size=w)

    def mousePressEvent(self, evt):
        self._separatorSelected = None
        x,y = evt.x, evt.y
        if self.border():
            x-=1 ; y-=1
        # TTkLog.debug(f"{self._separators} {evt}")
        for i, val in enumerate(self._separators):
            if self._orientation == TTkK.HORIZONTAL:
                if x == val:
                    self._separatorSelected = i
                    self.update()
                    self._updateGeometries()
            else:
                if y == val:
                    self._separatorSelected = i
                    self.update()
                    self._updateGeometries()
        return self._separatorSelected is not None

    def mouseDragEvent(self, evt):
        if self._separatorSelected is not None:
            x,y = evt.x, evt.y
            if self.border():
                x-=1 ; y-=1
            if self._orientation == TTkK.HORIZONTAL:
                self._separators[self._separatorSelected] = x
            else:
                self._separators[self._separatorSelected] = y
            self._updateGeometries()
            self.update()
            return True
        return False

    def focusOutEvent(self):
        self._separatorSelected = None

    def minimumHeight(self) -> int:
        ret = 2 if self.border() else 0
        if not self._splitterInitialized: return ret
        if self._orientation == TTkK.VERTICAL:
            for item in self._items:
                ret+=item.minimumHeight()+1
            ret = max(0,ret-1)
        else:
            for item in self._items:
                if ret < item.minimumHeight():
                    ret = item.minimumHeight()
        return ret

    def minimumWidth(self)  -> int:
        ret = 2 if self.border() else 0
        if not self._splitterInitialized: return ret
        if self._orientation == TTkK.HORIZONTAL:
            for item in self._items:
                ret+=item.minimumWidth()+1
            ret = max(0,ret-1)
        else:
            for item in self._items:
                if ret < item.minimumWidth():
                    ret = item.minimumWidth()
        return ret

    def maximumHeight(self) -> int:
        b = 2 if self.border() else 0
        if not self._splitterInitialized: return 0x10000
        if self._orientation == TTkK.VERTICAL:
            ret = b
            for item in self._items:
                ret+=item.maximumHeight()+1
            ret = max(b,ret-1)
        else:
            ret = 0x10000
            for item in self._items:
                if ret > item.maximumHeight():
                    ret = item.maximumHeight()
        return ret

    def maximumWidth(self)  -> int:
        b = 2 if self.border() else 0
        if not self._splitterInitialized: return 0x10000
        if self._orientation == TTkK.HORIZONTAL:
            ret = b
            for item in self._items:
                ret+=item.maximumHeight()+1
            ret = max(b,ret-1)
        else:
            ret = 0x10000
            for item in self._items:
                if ret > item.maximumWidth():
                    ret = item.maximumWidth()
        return ret

    _ttkProperties = {
        'Border' : {
                'init': {'name':'border', 'type':bool } },
        'Orientation' : {
                'init': {'name':'orientation', 'type':'singleflag',
                    'flags': {
                        'Horizontal' : TTkK.HORIZONTAL ,
                        'Vertical'   : TTkK.VERTICAL   } } ,
                'get':  {'cb':orientation,     'type':'singleflag',
                    'flags': {
                        'Horizontal' : TTkK.HORIZONTAL ,
                        'Vertical'   : TTkK.VERTICAL   } } }
    }
